from remotemanager.storage.function import Function
from remotemanager.logging import LoggingMixin


class RemoteFunction(LoggingMixin):
    """
    Decorator class to store a function within the "cached functions" property.

    Wrap a function with `@RemoteFunction` to store:

    >>> @RemoteFunction
    >>> def func(val):
    >>>     return val

    This function will then be made available in all runs:

    >>> def main():
    >>>     val = ...
    >>>     return func(val)
    >>>
    >>> ds = Dataset(main)
    >>> ...

    .. versionadded:: 0.3.6
    """

    def __init__(self, function):

        self._storedfunction = function

        storedfunction = Function(function)
        self._logger.info(f'caching function {storedfunction.raw_source}')
        cached_functions[function.__name__] = storedfunction

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    @property
    def function(self):
        return self._storedfunction


cached_functions = {}



import logging
import os.path
import warnings

from .utils import format_iterable


class Handler:

    """
    Handler class for the global logging. Allows for limited user control
    """

    _levels = {'CRITICAL': logging.CRITICAL,
               'ERROR': logging.ERROR,
               'WARNING': logging.WARNING,
               'INFO': logging.INFO,
               'DEBUG': logging.DEBUG}

    _filetype = '.yaml'
    _default_file = 'remotemanager_log'
    _defaultpath = f'{_default_file}{_filetype}'

    def __init__(self):
        self._base = logging.getLogger('remotemanager')

        self._external = logging.getLogger('remotemanager.EXTERNAL')
        self._external.setLevel(10)  # external calls should always be logged

        self._path = Handler._defaultpath
        self._mode = 'w'
        self._level = 'WARNING'  # default to WARNING

        self._base.setLevel(Handler._levels[self._level])

        self._update_handlers()

    def refresh(self):
        """
        deletes and re-creates the log
        """
        self._delete_log(warn=False,
                         force=True)
        self._update_handlers()

    @property
    def level(self):
        """
        Return the string format of the current logging level
        """
        return self._level

    @property
    def path(self):
        """
        Attribute determining the current log path
        """
        return self._path

    @property
    def write_mode(self):
        """
        returns the current write-mode of the logger
        """
        return self._mode

    @write_mode.setter
    def write_mode(self, mode):
        self._mode = mode
        self._update_handlers()

    @property
    def overwrite(self):
        return 'w' in self._mode

    @overwrite.setter
    def overwrite(self, mode):
        if mode:
            self._mode = 'w'
        else:
            self._mode = 'a'
        self._update_handlers()

    @property
    def empty(self):
        return _check_log_empty(self._path)

    @level.setter
    def level(self, level):
        f"""
        Sets the logging level to `level`

        Arguments:
            level (str):
                Update the logging level. See the _levels attribute for 
                valid options
        """
        level = level.upper()
        if level not in Handler._levels.keys():
            raise ValueError('log level must be one of '
                             f'{list(Handler._levels.keys())}')
        self._base.setLevel(Handler._levels[level])
        self._level = level

    @path.setter
    def path(self, newpath):
        """
        set logging path to newpath, deleting old log if it is empty

        Arguments:
            newpath (str):
                new path to log at
        """
        if newpath is None:
            newpath = ''
        path, file = os.path.split(newpath)
        if file == '':
            file = Handler._defaultpath
        if path == '':
            path = '.'

        newpath = os.path.join(path, file)
        if Handler._filetype not in newpath:
            path, file = os.path.split(newpath)
            newpath = os.path.join(path, f'{file}{Handler._filetype}')

        if not os.path.isdir(path):
            os.makedirs(path)

        self._delete_log(warn=True)

        self._path = newpath
        self._update_handlers()

    def _delete_log(self,
                    warn: bool = False,
                    force: bool = False):
        """
        Deletes the current log

        Arguments:
            warn (bool):
                warn the user if the log cannot be deleted
            force (bool):
                always delete
        """
        def delete_log(path):
            try:
                os.remove(path)
            except PermissionError:
                warnings.warn(f'could not remove log at {self._path} '
                              f'(file open)')

        if _check_log_empty(self._path):
            delete_log(self._path)
        elif force:
            delete_log(self._path)
        elif warn:
            warnings.warn(f'could not remove log at {self._path} (not empty)')

    def _update_handlers(self):
        """
        refresh handlers attached to the logger
        """
        for handler in self._base.handlers:
            self._base.removeHandler(handler)
        file_handler = logging.FileHandler(self._path, mode=self._mode)
        formatter = logging.Formatter('%(asctime)s - '
                                      '%(levelname)s - '
                                      '%(name)s.%(funcName)s: '
                                      '%(message)s',
                                      datefmt='%Y-%m-%d %H-%M-%S')

        file_handler.setFormatter(formatter)
        self._base.addHandler(file_handler)
    # TODO(lbeal) can this be made more pythonic?

    def debug(self, msg, *args, **kwargs):
        """Direct passthrough for `debug` logging method"""
        if not isinstance(msg, str):
            msg = format_iterable(msg)
        self._external.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """Direct passthrough for `info` logging method"""
        if not isinstance(msg, str):
            msg = format_iterable(msg)
        self._external.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """Direct passthrough for `warning` logging method"""
        if not isinstance(msg, str):
            msg = format_iterable(msg)
        self._external.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """Direct passthrough for `error` logging method"""
        if not isinstance(msg, str):
            msg = format_iterable(msg)
        self._external.error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """Direct passthrough for `critical` logging method"""
        if not isinstance(msg, str):
            msg = format_iterable(msg)
        self._external.critical(msg, *args, **kwargs)


def _check_log_empty(path):
    """
    check that the log at path `path` is clean
    """
    with open(path, 'r') as o:
        log = o.read()
    if log.strip() == '':
        return True
    return False



from remotemanager.connection.computers.base import BaseComputer, \
    required, optional
from remotemanager.utils import safe_divide


class Vega(BaseComputer):
    """
    example class for connecting to a remote computer

    initial info used from vega docs
    https://en-vegadocs.vega.izum.si/slurm/

    completed using vega.py from BigDFT, 221114
    """

    def __init__(self, **kwargs):

        # default host to vega's IP
        if 'host' not in kwargs:
            kwargs['host'] = 'login.vega.izum.si'

        # initialise the base URL
        super().__init__(**kwargs)

        # script housekeeping
        self.submitter = 'sbatch'
        self.shebang = '#!/bin/bash -l'
        self.pragma = '#SBATCH'

        # resources
        self.mpi = required('--ntasks')
        self.omp = required('--cpus-per-task')
        self.time = required('--time')

        self.nodes = optional('--nodes')
        self.jobname = optional('--job-name', 'remote_run')
        self.outfile = optional('--output')
        self.errfile = optional('--error')
        self.gpus = optional('--gres gpus')
        self.queue = optional('--queue')
        self.reservation = optional('--reservation')

        # modules
        self.modules = ['Python/3.10.4-GCCcore-11.3.0']
        self.modules_gpu = []

    @property
    def cpu_node(self):
        return {'cores_per_node': 256,
                'partition': 'cpu'}

    @property
    def gpu_node(self):
        return {'cores_per_node': 128,
                'gpus_per_node': 4,
                'partition': 'gpu'}

    def resources_block(self, **kwargs):
        self.update_resources(**kwargs)

        self.enable_gpu = kwargs.get('gpu', False)

        if not self.enable_gpu and self.gpus:
            self.enable_gpu = True

        if not self.valid:
            raise RuntimeError(f'missing required arguments: {self.missing}')

        node_specs = self.gpu_node if self.enable_gpu else self.cpu_node

        cores_per_node = node_specs['cores_per_node']
        nodes = safe_divide(self.mpi.value * self.omp.value, cores_per_node)

        ### safeguards ###
        # limit OMP to cores_per_machine
        if self.omp.value > cores_per_node:
            raise ValueError(f'OMP={self.omp.value} is larger than the number of '
                             f'cores per node ({cores_per_node})')
        # limit GPU request to 4 per node
        if self.gpus and self.gpus.value > node_specs['gpus_per_node']:
            raise ValueError(f'gpus={self.gpus.value} is larger than the number of '
                             f'gpus per node ({node_specs["gpus_per_node"]})')

        output = []
        for k, v in self.argument_dict.items():
            if k in ['nodes', 'gpus']:
                continue

            if v:
                output.append(f'{self.pragma} {v.flag}={v.value}')

        if self.enable_gpu:
            gpus = self.gpus.value if self.gpus else node_specs["gpus_per_node"]

            output.append(f'{self.pragma} --gres=gpu:{gpus * nodes}')
            output.append(f'{self.pragma} --partition=gpu')

        output.append(f'{self.pragma} --nodes={nodes}')

        return output

    def modules_block(self):
        if self.enable_gpu:
            modules = self.modules_gpu
        else:
            modules = self.modules
        return ['\nmodule purge'] + [f'module load {m}' for m in modules]

    def script(self,
               **kwargs) -> str:
        """
        Takes job arguments and produces a valid jobscript

        Returns:
            (str):
                script
        """
        self.enable_gpu = False  # default off
        script = [self.shebang]

        script += self.resources_block(**kwargs)
        script += self.modules_block()

        if self.enable_gpu:
            try:
                script.append(self.extra_gpu)
            except AttributeError:
                script.append(self.extra)

        elif hasattr(self, 'extra') and self.extra is not None:
            script.append(self.extra)

        return '\n'.join(script)



"""
URL base class for connecting to remote systems
"""
import logging
import os
import re
import typing

from remotemanager.logging.utils import format_iterable
from remotemanager.connection.cmd import CMD
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils import ensure_list
from remotemanager.logging import LoggingMixin


class URL(SendableMixin, LoggingMixin):
    """
    Container to store the url info for a Remote run

    The url should contain everything pertaining to the _remote_, allowing
    Dataset to be remote-agnostic

    Arguments:
        user (str):
            username for the remote system
        host (str):
            host address of the remote system
        port (int, str):
            port to connect to for ssh tunnels
        verbose (bool):
            base-verbosity for connections
        timeout (int):
            time to wait before issuing a timeout for cmd calls
        max_timeouts (int):
            number of times to attempt cmd communication in case of a timeout
        python (str):
            string used to initiate a python instance
        raise_errors (bool):
            set false to ignore errors by default in cmd calls
        passfile (str):
            absolute path to password file for sshpass
        kwargs:
            any extra args that may end up here from a Dataset or Computer are
            discarded
    """

    _localhost = 'localhost'

    def __init__(self,
                 user: str = None,
                 host: str = None,
                 port: int = None,
                 verbose: bool = False,
                 timeout: int = 5,
                 max_timeouts: int = 3,
                 python: str = 'python',
                 raise_errors: bool = True,
                 passfile: str = None,
                 **kwargs):

        if host is None:
            host = URL._localhost
        self._conn = {'user': user,
                      'host': host,
                      'port': port}

        self.timeout = timeout
        self.max_timeouts = max_timeouts
        self.python = python

        self._passfile = passfile

        self._verbose = verbose
        self._raise_errors = raise_errors
        self._ssh_override = None

        self._logger.info('new url created with url details:'
                          f'{format_iterable(self._conn)}')

    @property
    def raise_errors(self):
        return self._raise_errors

    @raise_errors.setter
    def raise_errors(self, r):
        self._raise_errors = r

    @property
    def user(self) -> str:
        """
        Currently configured username
        """
        return self._conn['user']

    @user.setter
    def user(self, user):
        """
        Set the user attribute
        """
        self._conn['user'] = user

    @property
    def host(self) -> str:
        """
        Currently configured hostname
        """
        return self._conn['host'] or URL._localhost

    @host.setter
    def host(self, host):
        """
        Set the host attribute
        """
        self._conn['host'] = host

    @property
    def userhost(self) -> str:
        """
        `user@host` string if possible, just `host` if user is not present
        """
        if self.user is None:
            return self.host
        else:
            return f'{self.user}@{self.host}'

    @property
    def port(self) -> int:
        """
        Currently configured port (defaults to 22)
        """
        port = self._conn['port'] or 22
        return port

    @property
    def ssh(self) -> str:
        """
        ssh insert for commands on this connection
        """
        if self.is_local:
            raise ValueError('ssh cannot be generated for a local url')

        ret = []

        if self.passfile is not None:
            ret.append(f'sshpass -f {self.passfile}')

        if not self._ssh_override:
            ret.append(f'ssh -p {self.port}')
        else:
            return self._ssh_override

        ret.append(self.userhost)

        return ' '.join(ret)

    @ssh.setter
    def ssh(self,
            newssh: str) -> None:
        """
        Allows forced override of the ssh command

        Inserting extra flags into the ssh can be done as follows:

        >>> url = URL()
        >>> print(url.ssh)
        >>> "ssh"
        >>> url.ssh = "LANG=C " + url.ssh
        >>> print(url.ssh)
        >>> "LANG=C ssh"

        Args:
            newssh (str):
                new ssh string to insert

        Returns:
            None
        """
        self._ssh_override = newssh

    def clear_ssh_override(self):
        """
        Wipe any override applied to ssh. Can also be done by setting
        url.ssh = None

        Returns:
            None
        """
        self._ssh_override = None

    @property
    def passfile(self):
        if self._passfile is None:
            return

        p = self._passfile.replace('~', os.environ['HOME'])
        if not os.path.isfile(p):
            raise RuntimeError(f'could not find password file at {self._passfile}')

        return self._passfile

    @property
    def is_tunnel(self):
        """
        Uses the presence of an assigned port to determine if we are tunnelling
        or not

        Returns (bool):
           True if this is a tunnelled connection
        """
        return self._conn['port'] is not None

    @property
    def is_local(self):
        """
        True if this connection is purely local
        """
        host = self.host
        if host == URL._localhost:
            return True
        elif host.startswith('127.'):
            return True
        return False

    def cmd(self,
            cmd: str,
            asynchronous: bool = False,
            local: [bool, None] = None,
            stdout: str = None,
            stderr: str = None,
            timeout: int = None,
            max_timeouts: int = None,
            raise_errors: bool = None,
            dry_run: bool = False) -> typing.Union[CMD, str]:
        """
        Creates and executes a command

        Args:
            asynchronous (bool):
                run this command asynchronously
            cmd (str):
                command to execute
            local (bool, None):
                force a local or remote execution. Defaults to None
            stdout (str):
                optional file to redirect stdout to
            stderr (str):
                optional file to redirect stderr to
            timeout (int):
                time to wait before issuing a timeout
            max_timeouts (int):
                number of times to attempt communication in case of a timeout
            raise_errors (bool):
                override for global setting. Raise any stderr if encountered
            dry_run (bool):
                don't exec the command if True, just returns the string

        Returns (CMD):
            returned command instance
        """
        if local is not None and not local:
            self._logger.info('forced remote call; appending ssh to cmd')
            cmd = f'{self.ssh} "{cmd}"'
        elif local is None and not self.is_local:
            self._logger.info('implicit remote call; appending ssh to cmd')
            cmd = f'{self.ssh} "{cmd}"'
        if raise_errors is None:
            raise_errors = self._raise_errors

        timeout = self.timeout \
            if timeout is None else timeout
        max_timeouts = self.max_timeouts \
            if max_timeouts is None else max_timeouts

        if dry_run:
            return cmd.strip()

        thiscmd = CMD(cmd.strip(),
                      asynchronous=asynchronous,
                      stdout=stdout,
                      stderr=stderr,
                      timeout=timeout,
                      max_timeouts=max_timeouts,
                      raise_errors=raise_errors)
        thiscmd.exec()

        return thiscmd

    @property
    def utils(self):
        """
        Handle for the URLUtils module
        """
        if not hasattr(self, '_urlutils'):
            self._urlutils = URLUtils(self)
        return self._urlutils


class URLUtils:
    """
    Extra functions to go with the URL class, called via URL.utils

    As it requires a parent `URL` to function, and is instantiated with a
    `URL`, there is little to no purpose to using this class exclusively

    Arguments:
        parent (URL):
            parent class to provide utils to
    """
    def __init__(self, parent: URL):

        self._logger = logging.getLogger(__name__ + '.URLUtils')
        self._logger.info(f'creating a utils extension to parent: {parent}')

        self._parent = parent

    def file_mtime(self,
                   files: list,
                   local: bool = None,
                   python: bool = False) -> dict:
        """
        Check file modification times of [files]

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search
            python (bool):
                ensure python style search is used

        Returns (dict):
            {file: mtime (unix)} dictionary
        """

        self._logger.info(f'performing stat on files: {files}')
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)
        times, error = self._file_mtime(files, local, python)
        self._logger.info('received:')
        self._logger.info(times)
        self._logger.info(error)
        output = {}
        idx = 0
        for file in files:
            if file in error:
                output[file] = None
            else:
                output[file] = times[idx]
                idx += 1

        return output

    def _file_mtime(self,
                    files: list,
                    local: bool = None,
                    python: bool = False):
        """
        Perform the "stat -c %Y" command on a list of files,
        returning the result. Uses a python command backup if this fails

        Args:
            files (list):
                list of files to check
            local (bool):
                force a local search
            python (bool):
                force the python override

        Returns:
            (list): list of file unix times
        """
        def regex_stat_error(stderr):
            search = re.search("\'(.*)\'", stderr)
            if search is None:
                return None
            return search.group(0).strip("'")

        def stat():
            self._logger.debug('attempting raw stat command on files')
            basecmd = 'stat -c %Y'
            if len(files) == 1:
                cmd = f'{basecmd} {files[0]}'
            else:
                cmd = f'{basecmd} {{' + ','.join(files) + '}'

            ret = self._parent.cmd(cmd,
                                   local=local,
                                   raise_errors=False)

            times = [int(t) for t in ret.stdout.split('\n') if t != '']
            error = [regex_stat_error(err) for err in ret.stderr.split('\n')]

            return times, error, ret.returncode, ret.stderr

        def pystat():
            self._logger.debug('attempting python stat on files')
            ex = f"""import os
files={files}
for f in files:
\ttry: print(os.stat(f).st_mtime) 
\texcept FileNotFoundError: print(f)"""

            cmd = f'{self._parent.python} -c "{ex}"'

            ret = self._parent.cmd(cmd,
                                   local=local,
                                   raise_errors=False)

            times = []
            error = []
            for line in ret.stdout.split('\n'):
                try:
                    times.append(int(float(line.strip())))
                except ValueError as E:
                    if "could not convert string to float" in str(E):
                        error.append(line.strip())

            return times, error

        files = ensure_list(files)

        if not python:
            t, e, returncode, stderr = stat()
            if returncode in [126, 127] or 'illegal option' in stderr:
                self._logger.warning('stat failed, falling back on python')
                return pystat()

            return t, e

        return pystat()

    def file_presence(self,
                      files: list,
                      local: bool = None) -> dict:
        """
        Search for a list of files, returning a boolean presence dict

        Args:
            files (list):
                list of paths to files
            local (bool):
                force a local search

        Returns (dict):
            {file: present} dictionary
        """
        self._logger.info(f'checking for presence of files: {files}')
        if local is None:
            local = self._parent.is_local

        files = ensure_list(files)

        times = self.file_mtime(files,
                                local=local)

        return {f: times[f] is not None for f in files}

    def search_folder(self,
                      files: list,
                      folder: str,
                      local: bool = None) -> dict:
        """
        Search `folder` for `files`, returning a boolean presence dict

        Arguments:
            files (list):
                list of filenames to check for. Optionally, a string for a
                single file
            folder (str):
                folder to scan
            local (bool):
                perform the scan locally (or remotely)

        Returns (dict):
            {file: present} dictionary
        """
        if local is None:
            local = self._parent.is_local
        fpath = os.path.abspath(folder)  # not locally available ?

        self._logger.debug(f'scanning folder {fpath}')
        self._logger.debug('searching for files:')
        self._logger.debug(f'{format_iterable(files)}')

        ls_return = self.ls(fpath, local=local, as_list=True)

        scan = [os.path.basename(f) for f in ls_return]

        self._logger.debug('scan sees:')
        self._logger.debug(f'{format_iterable(scan)}')

        if isinstance(files, str):
            self._logger.info('files is a string, running in singular mode')
            ret = {files: os.path.basename(files) in scan}
        else:
            ret = {file: os.path.basename(file) in scan for file in files}

        return ret

    def touch(self,
              file: str,
              local: bool = None,
              raise_errors: bool = None) -> CMD:
        """
        perform unix `touch`, creating or updating `file`

        Arguments:
            file (str):
                filename or path to file
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f'utils touch on file {file}')
        fname = os.path.abspath(file)
        return self._parent.cmd(f'touch {fname}',
                                local=local,
                                raise_errors=raise_errors)

    def mkdir(self,
              file: str,
              local: bool = None,
              raise_errors: bool = None) -> CMD:
        """
        perform unix `mkdir -p`, creating a folder structure

        Arguments:
            file (str):
                name or path to folder
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered

        Returns (CMD):
            CMD instance for the command
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f'utils mkdir on path {file}')
        fname = os.path.abspath(file)
        # print(f'making dir {fname}')
        return self._parent.cmd(f'mkdir -p {fname}',
                                local=local,
                                raise_errors=raise_errors)

    def ls(self,
           file: str,
           as_list: bool = True,
           local: bool = None,
           raise_errors: bool = None) -> [CMD, list]:
        """
        Identify the files present on the directory

        Arguments:
            file (str):
                name or path to folder.
            as_list (bool):
                convert to a list format
            local (bool):
                force local (or remote) execution
            raise_errors (bool):
                raise any stderr encountered

        Returns (CMD, list):
            CMD instance for the command, or the list if as_list is True
        """
        if local is None:
            local = self._parent.is_local
        self._logger.debug(f'utils ls on path {file}')
        fname = os.path.abspath(file) if local else file

        ret = self._parent.cmd(f'ls {fname}',
                               local=local,
                               raise_errors=raise_errors)

        if as_list:
            ret = [f for f in ret.stdout.split('\n') if f != '']
        return ret



from remotemanager.connection.computers.summer import Summer
from remotemanager.connection.computers.vega import Vega
from remotemanager.connection.computers.base import BaseComputer, \
    optional, required

__all__ = ['BaseComputer',
           'optional',
           'required',
           'Summer',
           'Vega']



import importlib
import inspect
import os
import typing

from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.utils.uuid import generate_uuid
from remotemanager.logging import LoggingMixin


_SCRIPT_TEMPLATE = """
{function}

if __name__ == '__main__':    
\tkwargs = {args}
    
\tresult = {name}(**kwargs)
"""


class Function(SendableMixin, LoggingMixin):
    """
    Serialise function to an executable python file for transfer

    Args:
        func:
            python function for serialisation
    """
    def __init__(self,
                 func: typing.Callable):

        self._logger.debug(f'creating new serialisable function for {func}')

        self._uuid = ''
        try:
            source = inspect.getsource(func)
            source = source.replace('@RemoteFunction', '')
            self._source = source
            self._fname = func.__name__
        except TypeError as e:
            if isinstance(func, str):
                self._source = func
                self._fname = "f"
            else:
                raise e
        self._uuid = generate_uuid(self._source)

    def __call__(self, *args, **kwargs):
        return self.object(*args, **kwargs)

    @property
    def name(self):
        """
        Function name
        """
        return self._fname

    @property
    def raw_source(self):
        """
        Function source
        """
        return self._source

    @property
    def uuid(self):
        """
        Function uuid (64 characters)
        """
        return self._uuid

    @property
    def object(self):
        """
        Recreates the function object by writing out the source, and importing.

        Returns:
            typing.Callable:
                the originally passed function
        """

        tmp_file = ''
        try:
            tmp_file = os.path.abspath(f'{self.uuid}.py')

            with open(tmp_file, 'w+') as o:
                o.write(self.raw_source)

            func_module = importlib.import_module(self.uuid)
            func_object = getattr(func_module, f'{self.name}')

        finally:
            os.remove(tmp_file)

        return func_object

    def dump_to_string(self, args):
        """
        Dump this function to a serialised string, ready to be written to a
        python file

        Args:
            args (dict):
                arguments to be used for this dumped run

        Returns:
            (str):
                serialised file
        """

        if args is None:
            args = {}

        return _SCRIPT_TEMPLATE.format(**{'function': self.raw_source,
                                          'name': self.name,
                                          'args': args})



"""
Baseclass for any file transfer
"""
import os.path

from remotemanager.connection.cmd import CMD
from remotemanager.connection.url import URL
from remotemanager.storage.sendablemixin import SendableMixin
from remotemanager.logging.utils import format_iterable
from remotemanager.utils import ensure_list, ensure_dir
from remotemanager.logging import LoggingMixin
from remotemanager.utils.flags import Flags


class Transport(SendableMixin, LoggingMixin):
    """
    Baseclass for file transfer

    Args:
        url (URL):
            url to extract remote address from
    """

    def __init__(self,
                 url: URL = None,
                 *args,
                 **kwargs):

        self._remote_address = None
        self._url = None
        self.set_remote(url)

        self._flags = Flags()
        self._transfers = {}

    def queue_for_push(self,
                       files: list,
                       local: str = None,
                       remote: str = None):
        """
        Queue file(s) for sending (pushing)

        Args:
            files (list[str], str):
                list of files (or file) to add to push queue
            local (str):
                local/origin folder for the file(s)
            remote (str):
                remote/destination folder for the file(s)
        Returns:
            None
        """
        self._logger.info(f'adding to PUSH queue')
        self.add_transfer(files, local, remote, 'push')

    def queue_for_pull(self,
                       files: list,
                       local: str = None,
                       remote: str = None):
        """
        Queue file(s) for retrieving (pulling)

        Args:
            files (list[str], str):
                list of files (or file) to add to pull queue
            local (str):
                local/destination folder for the file(s)
            remote (str):
                remote/origin folder for the file(s)
        Returns:
            None
        """
        self._logger.info(f'adding to PULL queue')
        self.add_transfer(files, remote, local, 'pull')

    def add_transfer(self,
                     files: list,
                     origin: str,
                     target: str,
                     mode: str):
        """
        Create a transfer to be executed. The ordering of the origin/target
        files should be considered as this transport instance being a
        "tunnel" between wherever it is executed (origin), and the destination
        (target)

        Args:
            files (list[str], str):
                list of files (or file) to add to pull queue
            origin (str):
                origin folder for the file(s)
            target (str):
                target folder for the file(s)
            mode (str: "push" or "pull"):
                transfer mode. Chooses where the remote address is placed
        Returns:
            None
        """
        modes = ('push', 'pull')
        if mode.lower() not in modes:
            raise ValueError(f'mode must be one of {modes}')

        if origin is None:
            origin = '.'
        if target is None:
            target = '.'

        # ensure dir-type
        origin = os.path.join(origin, '')
        target = os.path.join(target, '')

        if mode == 'push':
            pair = f'{origin}>{self._add_address(target)}'
        else:
            pair = f'{self._add_address(origin)}>{target}'

        files = [os.path.split(f)[1] for f in ensure_list(files)]

        self._logger.info(f'adding transfer: {split_pair(pair)[0]} '
                          f'-> {split_pair(pair)[1]}')
        self._logger.info(f'for files {files}')

        if pair in self._transfers:
            self._transfers[pair] = self._transfers[pair].union(set(files))
        else:
            self._transfers[pair] = set(files)

    def _add_address(self, dir: str) -> str:
        """
        Adds the remote address to the dir `dir` if it exists

        Args:
            dir (str):
                remote dir to have address appended

        Returns:
            (str) dir
        """
        if self.address is None:
            return dir
        return f'{self.address}:{dir}'

    @staticmethod
    def _format_for_cmd(folder: str, inp: list) -> str:
        """
        Formats a list into a bash expandable command with brace expansion

        Args:
            folder (str):
                the dir to copy to/from
            inp (list):
                list of items to compress

        Returns (str):
            formatted cmd
        """

        if isinstance(inp, str):
            raise ValueError('files is stringtype, '
                             'was a transfer forced into the queue?')

        if len(inp) > 1:
            return os.path.join(folder, '{' + ','.join(inp) + '}')
        return os.path.join(folder, inp[0])

    @property
    def transfers(self) -> dict:
        """
        Return the current transfer dict

        Returns (dict):
            {paths: files} transfer dict
        """
        return {k: sorted(list(v)) for k, v in self._transfers.items()}

    def print_transfers(self):
        """
        Print a formatted version of the current queued transfers

        Returns:
            None
        """
        i = 0
        for pair, files in self.transfers.items():
            i += 1
            print(f'transfer {i}:'
                  f'\norigin: {split_pair(pair)[0]}'
                  f'\ntarget: {split_pair(pair)[1]}')
            j = 0
            for file in files:
                j += 1
                print(f'\t({j}/{len(files)}) {file}')

    @property
    def address(self):
        """
        return the remote address

        Returns (str):
            the remote address
        """
        return self._remote_address

    @address.setter
    def address(self, remote_address):
        """
        set the remote address

        Returns:
            None
        """
        self._remote_address = remote_address

    @property
    def url(self):
        if self._url is not None:
            return self._url
        return URL()

    @url.setter
    def url(self, url):
        self._url = url

    def set_remote(self,
                   url: URL = None):
        """
        set the remote address with a URL object

        Returns:
            None
        """
        if url is None:
            self._remote_address = None
        elif url.is_local:
            self._remote_address = None
        else:
            self._remote_address = url.userhost
            self.url = url

    @property
    def flags(self):
        return self._flags

    @flags.setter
    def flags(self, new):
        self._flags = Flags(str(new))

    @property
    def cmd(self):
        """
        Returns a semi formatted command for issuing transfers. It is left to
        the developer to implement this method when adding more transport
        classes.

        The implementation should ideally be an "intermediate" format between
        the stored `Transport._cmd` string and the final format.

        At its most basic:

        ```
        @property
        def cmd(self):
            base = self._cmd.format(primary='{primary}',
                                    secondary='{secondary}')
            return base
        ```

        where self._cmd is a string of the format

        `command {primary} {secondary}`

        This allows any further injection such as the flags that
        `Transport.rsync` uses.

        Returns (str):
            the semi-formatted command for issuing a transfer
        """
        raise NotImplementedError

    def transfer(self,
                 dry_run: bool = False):
        """
        Perform the actual transfer

        Args:
            dry_run (bool):
                do not perform command, just return the command(s) to be
                executed

        Returns (str, None):
            the dry run string, or None
        """
        def get_remote_dir(path):
            if ':' not in path:
                return ''
            return path.split(':')[1]

        self._logger.info(f'executing a transfer')

        commands = []
        for pair, files in self.transfers.items():

            primary, secondary = split_pair(pair)

            remote_dir = get_remote_dir(secondary)
            if remote_dir == '':
                remote_dir = get_remote_dir(primary)

            primary = self._format_for_cmd(primary, files)

            base_cmd = self.cmd.format(primary=primary,
                                       secondary=secondary,
                                       remote_dir=remote_dir)

            commands.append(base_cmd)

        if dry_run:
            return commands

        for cmd in commands:
            self.url.cmd(cmd, local=True)
            # wipe the transfer queue
            self.wipe_transfers()

    def wipe_transfers(self):
        self._logger.info('wiping transfers')
        self._transfers = {}


def split_pair(pair: str) -> list:
    """
    Convert a "dir>dir" string into list format

    Args:
        pair (tuple):
            (dir, dir) tuple to be split

    Returns (list):
        [dir, dir]

    """
    return [ensure_dir(os.path.split(p)[0]) for p in pair.split('>')]



"""
Handles file transfer via the `scp` protocol
"""
import os

from remotemanager.transport.transport import Transport
from remotemanager.utils import ensure_list


class scp(Transport):
    """
    Class to handle file transfers using the scp protocol

    Args:
        url (URL):
            url to extract remote address from
    """

    def __init__(self,
                 *args,
                 **kwargs):

        super().__init__(*args, **kwargs)

        # flags can be exposed, to utilise their flexibility
        flags = kwargs.pop('flags', 'r')
        self.flags = flags

        self._transfers = {}

        self._cmd = '{password}scp {flags} {primary} {secondary}'

    @staticmethod
    def _format_for_cmd(folder: str, inp: list) -> str:
        """
        Formats a list into a bash expandable command with brace expansion

        Args:
            folder (str):
                the dir to copy to/from
            inp (list):
                list of items to compress

        Returns (str):
            formatted cmd
        """
        def scp_join(files):
            return f' '.join([os.path.join(folder, f) for f in files])

        if isinstance(inp, str):
            raise ValueError('files is stringtype, '
                             'was a transfer forced into the queue?')

        inp = ensure_list(inp)

        if ':' not in folder:
            return scp_join(inp)
        remote, folder = folder.split(':')
        return f'{remote}:"{scp_join(inp)}"'

    @property
    def cmd(self):
        password = ''

        if self.url.passfile is not None:
            password = f'sshpass -f {self.url.passfile} '

        base = self._cmd.format(password=password,
                                flags=self.flags,
                                primary='{primary}',
                                secondary='{secondary}')
        self._logger.debug(f'returning semi-formatted cmd: "{base}"')
        return base
